from peacedeath import pm
import time


class Valves:
    def __init__(self, pe):
        self.pe = pe
        self.dev = pe.dev

    def _set_pwm(self, tube, pwm):
        """
        Sets the pwm signal that controls the rotor position of the servo valves.
        :param tube: tube index between 0 and 6
        :param pwm: Pulse Width Modulation period in microseconds
        """
        servo_remapping = {3: 0,
                           4: 1,
                           5: 2,
                           6: 3,
                           2: 4,
                           1: 5,
                           0: 6}
        tube = servo_remapping[tube]

        if tube == 0:
            self.dev.request(pm.dev_msg.SetServoStateRequest(servo0=pwm))
        elif tube == 1:
            self.dev.request(pm.dev_msg.SetServoStateRequest(servo1=pwm))
        elif tube == 2:
            self.dev.request(pm.dev_msg.SetServoStateRequest(servo2=pwm))
        elif tube == 3:
            self.dev.request(pm.dev_msg.SetServoStateRequest(servo3=pwm))
        elif tube == 4:
            self.dev.request(pm.dev_msg.SetServoStateRequest(servo4=pwm))
        elif tube == 5:
            self.dev.request(pm.dev_msg.SetServoStateRequest(servo5=pwm))
        elif tube == 6:
            self.dev.request(pm.dev_msg.SetServoStateRequest(servo6=pwm))

        time.sleep(0.3)  # wait for valve to move in desired position

        # cut power to all 7 valves
        self.dev.request(pm.dev_msg.SetServoStateRequest(servo0=0, servo1=0, servo2=0, servo3=0,
                                                         servo4=0, servo5=0, servo6=0))

    def open_valve(self, tube):
        self._set_pwm(tube, pwm=2300)

    def close_valve(self, tube):
        self._set_pwm(tube, pwm=500)

    def close_all_valves(self):
        for t in range(7):
            self.close_valve(tube=t)
            time.sleep(0.05)

    def open_all_valves(self):
        for t in range(7):
            self.open_valve(tube=t)




import numpy as np
import matplotlib.pyplot as plt
from peacedeath import pm
import os
import time

class Photodiodes:
    def __init__(self, pe):
        self.pe = pe
        self.dev = pe.dev
        self.log_file_current = os.path.join(self.pe.name, "log_current.csv")


    def measure_current(self, tube, cycles=10, n_measurements=40, laserOnDelay=10, laserOffDelay=10,
                        filter_errors=True, verbose=False, log = True):
        IFS, ISS, IFSoff, ISSoff = _measure_current(dev=self.dev, tube=tube, cycles=cycles,
                                                       n_measurements=n_measurements, laserOnDelay=laserOnDelay,
                                                       laserOffDelay=laserOffDelay, filter_errors=filter_errors,
                                                       verbose=verbose)
        IFS = round(IFS, 2)
        ISS = round(ISS, 2)
        IFSoff = round(IFSoff, 2)
        ISSoff = round(ISSoff, 2)
        if log:
            self.log_current(tube, IFS, ISS, IFSoff, ISSoff)
        return IFS, ISS, IFSoff, ISSoff

    def log_current(self, tube, IFS, ISS, IFSoff, ISSoff):
        if not os.path.exists(self.log_file_current):
            with open(self.log_file_current, "w+") as f:
                f.write("time, tube, front, side, front_off, side_off\n")
        current_time=time.ctime()
        with open(self.log_file_current, "a") as f:
            f.write("{time}, {tube:1d}, {IFS:.2f}, {ISS:.2f}, {IFSoff:.2f}, {ISSoff:.2f}\n".format(time=current_time,
                                                                                                  tube=tube,
                                                                                                  IFS=IFS,
                                                                                                  ISS=ISS,
                                                                                                  IFSoff=IFSoff,
                                                                                                  ISSoff=ISSoff))


def get_4_signals(dev, tube, resistor1=100, resistor2=10, laserOnDelay=10, laserOffDelay=1,
                  cycles=10, bias=0):
    tube_remapping = {0: 0,
                      1: 3,
                      2: 1,
                      3: 5,
                      4: 2,
                      5: 6,
                      6: 4}
    tube = tube_remapping[tube]
    assert 0 < cycles < 100
    s4 = dev.request(pm.dev_msg.ODMeasureRequest(tube=tube, laserOnDelay=laserOnDelay,
                                                 laserOffDelay=laserOffDelay,
                                                 cycles=cycles,
                                                 bias=bias,
                                                 resistor1=resistor1,
                                                 resistor2=resistor2))

    return s4.onFS, s4.onSS, s4.offFS, s4.offSS


def get_r2list(dev, tube=4, n=20):
    assert n > 5
    r1 = 200
    r2 = 128
    cycles = 1
    v = np.nan
    while not 0.9 < v < 1.1 and 5 < r2 < 255:
        a = get_4_signals(dev=dev, tube=tube, resistor1=r1, resistor2=r2, laserOnDelay=10, laserOffDelay=10,
                          cycles=cycles, bias=0)
        v = [(3.3 / 4095) * i / (10 * cycles) for i in a]
        v = v[0]  # onFS
        r2 /= (v - 0.1)
        r2 = min(255, r2)
        r2 = max(1, r2)
        r2 = int(r2)
    #         print(r2,v)

    rmax = min(int(r2 * 1.2), 255)
    rmax = max(rmax, 6)

    if r2 > 159 and v <= 0.2:
        rmax = 255
    if r2 > 159 and v > 0.2:
        rmax = 127
    r2list = list(range(0, rmax))
    if rmax > n:
        step = rmax // n
        step = max(1, step)
        r2list = r2list[1::step]
    return r2list


def measure_voltage(dev, tube, cycles=10, n_measurements=40, laserOnDelay=10, laserOffDelay=10,verbose=False):
    rlist = get_r2list(dev, tube, n=n_measurements)
    r1 = 200
    V = []
    Rfs = []
    Rss = []
    if len(rlist) < n_measurements:
        cycles *= n_measurements / len(rlist)
        cycles = int(cycles)
        if cycles > 100:
            cycles = 99
        if verbose:
            print("Cycles=", cycles)
    for r1_binary, r2_binary in zip(np.linspace(1, 255, len(rlist)).astype(int), rlist):
        #         print(r1_binary,r2_binary)
        u = get_4_signals(dev=dev, tube=tube, resistor1=r1_binary, resistor2=r2_binary, laserOnDelay=laserOnDelay,
                             laserOffDelay=laserOffDelay, cycles=cycles, bias=0)
        u = [(3.3 / 4095) * i / (10 * cycles) for i in u]
        V += [u]
        r2 = int(52 + r2_binary * 10000 / 256) # MCP42010 actual resistance
        r1 = int(52 + r1_binary * 10000 / 256) # MCP42010 actual resistance
        Rfs += [r2]
        Rss += [r1]
    Vfs_on = np.array(V)[:, 0]  # onFS
    Vss_on = np.array(V)[:, 1]  # onSS

    Vfs_off = np.array(V)[:, 2]  # offFS
    Vss_off = np.array(V)[:, 3]  # offSS
    Rfs = np.array(Rfs)
    Rss = np.array(Rss)
    return Vfs_on, Vfs_off, Rfs, Vss_on, Vss_off, Rss


def calculate_current(V, R, filter_errors=True):
    R = R[(V < 1.6)]
    V = V[(V < 1.6)]
    n = len(R)

    A = np.vstack((R, np.ones(n))).T
    fit = np.linalg.lstsq(A, V, rcond=None)
    current, voltage_bias = fit[0]
    resid = fit[1]
    rsquared = 1 - resid / (n * np.var(V))
    rsquared = rsquared[0]

    if filter_errors:
        #         print("R2=%.6f, I=%.2fuA, V_bias=%.3f unfiltered"%(rsquared,current*1e6,voltage_bias))
        err = (current * R + voltage_bias - V) ** 2
        valid_range = err < err.mean() + err.std()
        if any(valid_range):
            R = R[valid_range]
            V = V[valid_range]
            n = len(R)
            A = np.vstack((R, np.ones(n))).T
            fit = np.linalg.lstsq(A, V, rcond=None)
            current, voltage_bias = fit[0]
            resid = fit[1]
            rsquared = 1 - resid / (n * np.var(V))
    if n < 5:
        print("fit with less than 5 points!")
    current *= 1e6  # microAmps

    #     print("R2=%.6f, I=%.2fuA, V_bias=%.3f"%(rsquared,current,voltage_bias))
    return current, voltage_bias, rsquared


def _measure_current(dev, tube, cycles=10, n_measurements=40, laserOnDelay=10, laserOffDelay=10, filter_errors=True,
                     verbose=False):
    Vfs_on, Vfs_off, Rfs, Vss_on, Vss_off, Rss = measure_voltage(dev=dev, tube=tube, cycles=cycles, n_measurements=n_measurements,
                                         laserOnDelay=laserOnDelay, laserOffDelay=laserOffDelay, verbose=verbose)


    IFS, _, R2FS = calculate_current(Vfs_on, Rfs, filter_errors=filter_errors)
    IFSoff, _, R2FS = calculate_current(Vfs_off, Rfs, filter_errors=filter_errors)

    ISS, _, R1SS = calculate_current(Vss_on, Rss, filter_errors=filter_errors)
    ISSoff, _, R1SS = calculate_current(Vss_off, Rss, filter_errors=filter_errors)

    if verbose:
        fig, axs = plt.subplots(2, 1)
        axs[0].plot(Rfs, Vfs_on, "ro--", label="Front Detector laser on, I=%.1f μA" % IFS)
        axs[0].plot(Rfs, Vfs_off, "ro--", label="Front Detector laser off, I=%.1f μA" % IFSoff, alpha=0.3)
        axs[0].legend()
        axs[1].plot(Rss, Vss_on, "bo--", label="Side Detector laser on, I=%.1f μA" % ISS)
        axs[1].plot(Rss, Vss_off, "bo--", label="Side Detector laser off, I=%.1f μA" % ISSoff,alpha=0.3)
        axs[1].legend()
        plt.xlabel("Resistance (Ohms)")
        axs[1].set_ylabel("Voltage")
        axs[0].set_ylabel("Voltage")
        plt.show()
    return IFS, ISS, IFSoff, ISSoff



import pandas as pd
import numpy as np
import time
import os

class ODmeasurement:
    def __init__(self, pe, fit_calibration_functions=True):
        self.pe = pe
        self.dev = pe.dev
        self.calibration_file_front = os.path.join(self.pe.name, "calibration_data_od_front.csv")
        self.calibration_file_side = os.path.join(self.pe.name, "calibration_data_od_side.csv")
        self.calibration_file_front_off = os.path.join(self.pe.name, "calibration_data_od_front_off.csv")
        self.calibration_file_side_off = os.path.join(self.pe.name, "calibration_data_od_side_off.csv")
        self.od_file = os.path.join(self.pe.name, "log_OD.csv")

        self.calibration_functions = None
        if os.path.exists(self.calibration_file_front):
            df = pd.read_csv(self.calibration_file_front, index_col=0)
            if any(np.isnan(df.values.ravel())):
                print("OD calibration data incomplete.")
            else:
                if fit_calibration_functions:
                    self.fit_calibration_curves()
        else:
            print("OD calibration data not found.")

    def collect_calibration_data(self, calibration_OD_values=None, starting_index=0):
        if calibration_OD_values is None:
            od_list = input("Please input 7 od values used for calibration:")
            od_list = [float(od) for od in od_list.replace(" ", ",").split(",") if od]
            calibration_OD_values=od_list
        CurrentFront = pd.DataFrame(columns=[0, 1, 2, 3, 4, 5, 6], index=calibration_OD_values)
        CurrentSide = pd.DataFrame(columns=[0, 1, 2, 3, 4, 5, 6], index=calibration_OD_values)
        CurrentFrontOFF = pd.DataFrame(columns=[0, 1, 2, 3, 4, 5, 6], index=calibration_OD_values)
        CurrentSideOFF = pd.DataFrame(columns=[0, 1, 2, 3, 4, 5, 6], index=calibration_OD_values)
        if starting_index >0:
            CurrentFront = pd.read_csv(self.calibration_file_front, index_col=0)
            CurrentSide = pd.read_csv(self.calibration_file_side, index_col=0)
            CurrentFrontOFF = pd.read_csv(self.calibration_file_front_off, index_col=0)
            CurrentSideOFF = pd.read_csv(self.calibration_file_side_off, index_col=0)
            for df in [CurrentFront, CurrentSide, CurrentFrontOFF, CurrentSideOFF]:
                df.columns = df.columns.astype(int)
        for i in range(starting_index, 7):
            instructions = "Please arrange tubes as follows:\n"
            for t in range(7):
                instructions += "OD {OD} in tube {t}\n".format(OD=calibration_OD_values[(t - i) % 7], t=t)
            instructions += "ENTER to continue, input x to exit:"
            next_step = input(instructions)
            if next_step.lower() == "x":
                break
            else:
                for t in range(7):
                    self.pe.mix(100)
                    time.sleep(3)
                    self.pe.mix(0)
                    time.sleep(3)
                    IFS, ISS, IFSoff, ISSoff = self.pe.Photodiodes.measure_current(t)
                    CurrentFront.loc[calibration_OD_values[(t - i) % 7], t] = IFS
                    CurrentFrontOFF.loc[calibration_OD_values[(t - i) % 7], t] = IFSoff
                    CurrentSide.loc[calibration_OD_values[(t - i) % 7], t] = ISS
                    CurrentSideOFF.loc[calibration_OD_values[(t - i) % 7], t] = ISSoff

                    CurrentFront.to_csv(self.calibration_file_front)
                    CurrentSide.to_csv(self.calibration_file_side)
                    CurrentFrontOFF.to_csv(self.calibration_file_front_off)
                    CurrentSideOFF.to_csv(self.calibration_file_side_off)
                    try:
                        from IPython.display import clear_output
                        clear_output()
                    except:
                        pass

                    print(color.RED+color.BOLD+"Front Current:"+color.END)
                    print(color.RED+color.BOLD+str(CurrentFront)+color.END)
                    print(color.BLUE+color.BOLD+"Side Current:"+color.END)
                    print(color.BLUE+color.BOLD+str(CurrentSide)+color.END)
                    print(color.YELLOW+color.BOLD+"Front Current, laser OFF:"+color.END)
                    print(color.YELLOW+color.BOLD+str(CurrentFrontOFF)+color.END)
                    print(color.DARKCYAN+color.BOLD+"Side Current, laser OFF:"+color.END)
                    print(color.DARKCYAN+color.BOLD+str(CurrentSideOFF)+color.END)
            print("\t\tBatch completed.\n\n")
        print("\t\tDone collecting calibration data!")

    def fit_calibration_curves(self):
        df = pd.read_csv(self.calibration_file_front, index_col=0)
        df.columns = df.columns.astype(int)
        OD = np.array(df.index).ravel()
        calibration_functions = []
        for tube in range(7):
            current = df.loc[:, tube]
            function = fit_one_calibration_curve(current, OD)
            calibration_functions += [function]
        self.calibration_functions = calibration_functions

    def calculate_OD_from_calibration_function(self, FrontCurrent, tube):
        f = self.calibration_functions[tube]
        calculated_od = f(FrontCurrent)
        return calculated_od

    def measure_OD(self, tube,log=True,cycles=5, n_measurements=20, laserOnDelay=10, laserOffDelay=10,
                        filter_errors=True, verbose=False):
        if callable(self.calibration_functions[tube]):
            IFS, ISS, IFSoff, ISSoff = self.pe.Photodiodes.measure_current(tube, cycles=cycles,
                                                       n_measurements=n_measurements, laserOnDelay=laserOnDelay,
                                                       laserOffDelay=laserOffDelay, filter_errors=filter_errors,
                                                       verbose=verbose)
            measured_od = self.calculate_OD_from_calibration_function(IFS, tube)
            if log:
                self.log_od(tube, measured_od)
            return measured_od
        else:
            print("Please calibrate OD sensors")

    def log_od(self, tube, od_value):
        if not os.path.exists(self.od_file):
            with open(self.od_file, "w+") as f:
                f.write("time,0,1,2,3,4,5,6\n")
        string = [""] * 7
        string[tube] = str(round(od_value, 3))
        string=",".join([time.ctime()] + string)+"\n"
        with open(self.od_file, "a") as f:
            f.write(string)

class color:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'

def fit_one_calibration_curve(current, OD):
    x = np.log(current)
    y = np.log(OD + 1)

    coefficients = np.polyfit(x, y, 3)
    poly = np.poly1d(coefficients)

    def current_to_OD(current):
        x = np.log(current)
        y = poly(x)
        OD = np.exp(y) - 1
        return OD

    return current_to_OD

def show_calibration_curves(pe,show=True):
    import matplotlib.pyplot as plt
    df = pd.read_csv(os.path.join(pe.name, "calibration_data_od_front.csv"), index_col=0)
    df.columns = df.columns.astype(int)
    for t in range(7):
        current = df[t].values
        OD = df.index
        x = np.log(current)
        y = np.log(OD + 1)

        f = pe.OD_measurement.calibration_functions[t]
        x = np.linspace(10, 1700, 1690)
        plt.plot(x, f(x))
        plt.plot(df[t].values, df.index, "ro")
        # plt.yscale("log")
        plt.grid()
        if show:
            plt.show()


import time
import os
import pandas as pd
from peacedeath import pm

PUMP_DATA_FILENAME = "log_pumps.csv"
PUMP_CALIBRATION_FILENAME = "calibration_pumps.csv"


class Pumps:
    def __init__(self, pe):
        self.pe = pe  # peacedeath.Experiment object
        self.dev = pe.dev  # device
        self.logfile = os.path.join(pe.name, PUMP_DATA_FILENAME)
        self.calibration_file = os.path.join(pe.name, PUMP_CALIBRATION_FILENAME)
        self._default_pump_coefficients = {1: 215000, 2: 215000, 3: 215000}
        self.calibrated_pump_coefficients = {1: None, 2: None, 3: None}
        self.read_calibration_data()

    def pump_peace(self, tube, volume):
        self.pe.Valves.close_all_valves()  # redundant valve close for safety
        self.log_pump_operation(tube=tube, pump1Volume=volume, comment="started pumping")
        self.pe.Valves.open_valve(tube)
        time.sleep(0.5)
        steps = int(volume * self.pump_coefficients[1])
        self.dev.request(pm.dev_msg.PumpRequest(pump1Volume=steps))
        time.sleep(0.5)  # release all pressure in the tubes before closing the valve
        self.pe.Valves.close_valve(tube)
        self.log_pump_operation(tube=tube, pump1Volume=volume, comment="completed pumping")

    def pump_death(self, tube, volume):
        self.pe.Valves.close_all_valves()  # redundant valve close for safety
        self.log_pump_operation(tube=tube, pump2Volume=volume, comment="started pumping")
        self.pe.Valves.open_valve(tube)
        time.sleep(0.5)
        steps = int(volume * self.pump_coefficients[2])
        self.dev.request(pm.dev_msg.PumpRequest(pump2Volume=steps))
        time.sleep(0.5)  # release all pressure in the tubes before closing the valve
        self.pe.Valves.close_valve(tube)
        self.log_pump_operation(tube=tube, pump2Volume=volume, comment="completed pumping")

    def pump_vacuum(self, tube, volume):
        self.pe.Valves.close_all_valves()  # redundant valve close for safety
        self.log_pump_operation(tube=tube, pump3Volume=volume, comment="started pumping")
        self.pe.Valves.open_valve(tube)
        time.sleep(0.5)
        steps = int(volume * self.pump_coefficients[3])
        self.dev.request(pm.dev_msg.PumpRequest(pump3Volume=steps))
        time.sleep(0.5)  # release all pressure in the tubes before closing the valve
        self.pe.Valves.close_valve(tube)
        self.log_pump_operation(tube=tube, pump3Volume=volume, comment="completed pumping")

    def log_pump_operation(self, tube, pump1Volume=0, pump2Volume=0, pump3Volume=0,comment=""):
        if not os.path.exists(self.logfile):
            with open(self.logfile, "w+") as f:
                f.write("Time, tube, pump1Volume, pump2Volume, pump3Volume, comment\n")
        current_time = time.ctime()
        with open(self.logfile, "a") as f:
            f.write("{time}, {tube:1d}, {v1:.3f}, {v2:.3f}, {v3:.3f}, {comment}\n".format(time=current_time,
                                                                                tube=tube,
                                                                                v1=pump1Volume,
                                                                                v2=pump2Volume,
                                                                                v3=pump3Volume,
                                                                                comment=comment))

    def calibrate(self, tube=None, requested_volume=None, real_volume=None):
        """
        Calibrate the pumps by weighting the volume pumped in a vial using a scale.
        Place the vial on the scale, making sure that the silicone tubes will not affect the measurement.
        """
        pump_input_string = """Please select which pump to calibrate:
        1 - peace
        2 - death
        3 - vacuum
        input 1,2 or 3"""
        pump = int(input(pump_input_string))
        if tube is None:
            tube = int(input("Please select which tube to pump into: \n (0,1,2,3,4,5 or 6)"))
            assert tube in [0, 1, 2, 3, 4, 5, 6]

        if requested_volume is None:
            requested_volume = float(input("Please input volume to pump in ml: \n(max 20)"))
            assert 0 < requested_volume < 20

        if real_volume is None:
            print("Sending {vol}ml request to pump {p}".format(vol=requested_volume, p=pump))
            if pump == 1:
                self.pump_peace(tube=tube, volume=requested_volume)
            if pump == 2:
                self.pump_death(tube=tube, volume=requested_volume)
            if pump == 3:
                self.pump_vacuum(tube=tube, volume=requested_volume)

            real_volume = float(input("What was the real volume [ml]?"))
            self.pump_coefficients[pump] = int(self.pump_coefficients[pump] * requested_volume / real_volume)
            self.calibrated_pump_coefficients[pump] = self.pump_coefficients[pump]
            self.write_calibration_data()

    def write_calibration_data(self):
        pd.DataFrame(self.calibrated_pump_coefficients, index=[0]).to_csv(self.calibration_file, index=False)
        print("Wrote calibration data\n{coefs} to file\n{file}".format(coefs=self.calibrated_pump_coefficients,
                                                                       file=self.calibration_file))

    def read_calibration_data(self,pump_calibration_filepath=None):
        """
        read calibration of current experiment, or from another file
        """
        if pump_calibration_filepath is None:
            pump_calibration_filepath = self.calibration_file

        if not os.path.exists(pump_calibration_filepath):
            print("Pump calibration file not found.")
            print("Please calibrate pumps before running experiment!")
            self.pump_coefficients = self._default_pump_coefficients
        else:
            coefficients = pd.read_csv(pump_calibration_filepath).T[0]
            keys = coefficients.keys().astype(int)
            values = [int(i) if i > 0 else None for i in coefficients.values]
            self.calibrated_pump_coefficients = dict(zip(keys, values))

            print("Calibrated pump coefficients read from "+pump_calibration_filepath)
            print(self.calibrated_pump_coefficients)

            self.pump_coefficients = self.calibrated_pump_coefficients
            if None in self.calibrated_pump_coefficients.values():
                print("Please calibrate all pumps! Using some default values.")
                for k in self.pump_coefficients.keys():
                    if self.pump_coefficients[k] is None:
                        self.pump_coefficients[k] = self._default_pump_coefficients[k]


from peacedeath.valves import Valves
from peacedeath.photodiodes import Photodiodes
from peacedeath.od_measurement import ODmeasurement
from peacedeath.pumps import Pumps
from peacedeath import pm
from peacedeath.experiment import start_experiment, _write_parameters_OD_thresholds, _write_parameters_drug_doses, \
    _read_parameters_OD_thresholds, _read_parameters_drug_doses
import pandas as pd
import numpy as np
import os
import time
import threading


class Experiment:
    def __init__(self, name="NewExperiment", connect=True, device_serial="auto", fit_calibration_functions=True,
                 total_dilution_volume = 10, mixer_speed = 30, active_tubes = ()):
        """
        :param name: Experiment name with no spaces, e.g. starting date
        :param device_serial: board serial number or "auto" to auto-connect
        """
        assert " " not in name
        self.name = name
        if not os.path.exists(self.name):
            os.mkdir(self.name)
        if connect:
            self.dev = self.connect_device(device_serial=device_serial)
            self.Valves = Valves(self)
            self.Photodiodes = Photodiodes(self)
            self.Pumps = Pumps(self)
        else:
            self.dev = None
            self.hub = None
            print("Not connected to device")
        self.OD_measurement = ODmeasurement(self, fit_calibration_functions=fit_calibration_functions)
        self.active_tubes = active_tubes
        self.mixer_speed = mixer_speed
        self.total_dilution_volume = total_dilution_volume
        self.experiment_thread = None
        self.parameters_OD_thresholds = [0, 0, 0, 0, 0, 0, 0]
        self.parameters_drug_doses = [0, 0, 0, 0, 0, 0, 0]

    def mix(self, speed=70, period=3000):
        if 0 < speed < 30:
            speed = 30
        if speed > 100:
            speed = 100

        if speed > 0:
            state = 1
        else:
            state = 0

        pulseWidth = int(period * speed / 100)
        period = int(period)
        self.dev.request(pm.dev_msg.SetMixerStateRequest(state=state, period=period, pulseWidth=pulseWidth))

        # log
        try:
            with open(os.path.join(self.name, "log_mixer.csv"), "a") as f:
                f.write(time.ctime() + ",%d\n" % speed)
        except FileNotFoundError:
            with open(os.path.join(self.name, "log_mixer.csv"), "w+") as f:
                f.write("time, speed \n")
                f.write(time.ctime() + ",%d\n" % speed)

    def connect_device(self, device_serial="auto"):
        # hub = pm.PMHub.local()
        self.hub = pm.PMHub("tcp://localhost:5555")
        if device_serial == "auto":
            dev = self.hub.get_devices()[0]
            self.dev = dev
            return dev

    def test_device(self):
        print("Testing OD measurement...")
        for t in range(7):
            self.Photodiodes.measure_current(t, 1, 10)
        print("Testing mixers...")
        for speed in [30, 50, 100]:
            self.mix(speed)
            time.sleep(1)
        self.mix(0)
        print("Testing valves...")
        for t in range(7):
            self.Valves.open_valve(t)
            self.Valves.close_valve(t)
        print("Testing pumps...")
        self.Pumps.pump_peace(6, 0.1)
        self.Pumps.pump_death(6, 0.1)
        self.Pumps.pump_vacuum(6, 0.1)

    def start_experiment(self):
        if self.is_running():
            print("Experiment thread is already running! Use pe.stop_experiment() to stop.")
        else:
            self.experiment_thread = threading.Thread(target=start_experiment, args=(self,))
            self.experiment_thread.daemon = True
            self.experiment_thread.start()

    def is_running(self):
        try:
            return self.experiment_thread.is_alive()
        except AttributeError:
            return False


    def stop_experiment(self):
        with open(os.path.join(self.name, "stop.experiment"), "w+"):
            pass

    @property
    def parameters_OD_thresholds(self):
        return _read_parameters_OD_thresholds(self)
    @parameters_OD_thresholds.setter
    def parameters_OD_thresholds(self,thresholds):
        _write_parameters_OD_thresholds(self, thresholds)


    @property
    def parameters_drug_doses(self):
        return _read_parameters_drug_doses(self)
    @parameters_drug_doses.setter
    def parameters_drug_doses(self, drug_doses):
        _write_parameters_drug_doses(self, drug_doses)


import os
import time
import pandas as pd
import schedule
import numpy as np


def read_last_OD_values(pe):
    def strtoint(s):
        try:
            return float(s)
        except:
            return None

    df=pd.read_csv(pe.OD_measurement.od_file, index_col=0)
    lastvalues = [0,0,0,0,0,0,0]
    for t in range(7):
        try:
            ODvalues=list(strtoint(s) for s in df.iloc[:, t])
            ODvalues=[od for od in ODvalues if od]
            lastvalues[t]=ODvalues[-1]
        except:
            pass
    return np.array(lastvalues)


def make_dilution(pe, tube, death_volume=0, peace_volume=0, extra_vacuum=5):
    pe.mix(30)   # Mixer speed > 30 while pumping may lead to power failure!
    if death_volume>0:
        pe.Pumps.pump_death(tube, death_volume)
    if peace_volume>0:
        pe.Pumps.pump_peace(tube, peace_volume)
    pe.Pumps.pump_vacuum(tube, death_volume+peace_volume)
    pe.mix(0)
    pe.Pumps.pump_vacuum(tube, extra_vacuum)
    pe.mix(30)


def _write_parameters_drug_doses(pe, drug_doses=(0, 0, 0, 0, 0, 0, 0)):
    assert len(drug_doses) == 7
    with open(os.path.join(pe.name, "parameters_dose.csv"), "w+") as f:
        f.write(",".join([str(d) for d in drug_doses]))


def _read_parameters_drug_doses(pe):
    df = pd.read_csv(os.path.join(pe.name, "parameters_dose.csv"), header=None)
    drug_doses = np.array(df.values.ravel()).astype(float)
    pe.drug_doses = drug_doses
    return drug_doses


def _write_parameters_OD_thresholds(pe, OD_thresholds=(0, 0, 0, 0, 0, 0, 0)):
    OD_thresholds = np.array(OD_thresholds).astype(float)
    assert len(OD_thresholds) == 7
    with open(os.path.join(pe.name, "parameters_OD_thresholds.csv"), "w+") as f:
        f.write(",".join([str(thr) for thr in OD_thresholds]))


def _read_parameters_OD_thresholds(pe):
    OD_thresholds = pd.read_csv(os.path.join(pe.name, "parameters_OD_thresholds.csv"), header=None).values[0]
    OD_thresholds = np.array(OD_thresholds.ravel()).astype(float)
    pe.OD_thresholds = OD_thresholds
    return OD_thresholds


def dilute_if_necessary(pe, total_dilution_volume=10):
    OD_thresholds = pe.parameters_OD_thresholds
    drug_doses = pe.parameters_drug_doses

    death_volumes = total_dilution_volume * drug_doses/100
    peace_volumes = total_dilution_volume - death_volumes
    for tube in range(7):
        if OD_thresholds[tube] > 0.15 and pe.tubes_status_active[tube]:
            if pe.last_OD_values[tube] > OD_thresholds[tube]:
                make_dilution(pe, tube,
                              death_volume=death_volumes[tube],
                              peace_volume=peace_volumes[tube],
                              extra_vacuum=5)


def pause_mixers_and_measure_OD(pe, mixer_speed):
    pe.mix(mixer_speed)
    time.sleep(2)
    pe.mix(0)
    time.sleep(3)
    for t in range(7):
        pe.last_OD_values[t] = pe.OD_measurement.measure_OD(t)
    pe.mix(mixer_speed)

class color:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'


def start_experiment(pe):

    print("Starting Experiment loop")
    pe.tubes_status_active = [i in pe.active_tubes for i in range(7)]
    if len(pe.active_tubes) == 0:
        print("####### NO ACTIVE TUBES! #######")
    print("Active tubes: ", dict(zip(range(7), pe.tubes_status_active)))
    pe.mix(pe.mixer_speed)
    try:
        pe.last_OD_values = read_last_OD_values(pe)
    except FileNotFoundError:
        pe.last_OD_values = [0]*7
    print("Last OD values: ", dict(zip(range(7), pe.last_OD_values)))

    try:
        pe.drug_doses = read_parameters_drug_doses(pe)
    except FileNotFoundError:
        write_parameters_drug_doses(pe)
        read_parameters_drug_doses(pe)
    print("Death doses:", dict(zip(range(7), pe.drug_doses)))

    try:
        pe.OD_thresholds = read_parameters_OD_thresholds(pe)
    except FileNotFoundError:
        write_parameters_OD_thresholds(pe)
        read_parameters_OD_thresholds(pe)
    print("OD thresholds:", dict(zip(range(7), pe.OD_thresholds)))

    schedule.clear()
    schedule.every().minute.at(":00").do(pause_mixers_and_measure_OD, pe, mixer_speed=pe.mixer_speed)
    schedule.every().minute.at(":15").do(dilute_if_necessary, pe, total_dilution_volume=pe.total_dilution_volume)

    stop_file = os.path.join(pe.name, "stop.experiment")
    while not os.path.exists(stop_file):
        schedule.run_pending()
        time.sleep(1)
    os.remove(stop_file)
    print("Experiment stopped")
